#include <list>
#include <windows.h>
#include <memory.h>
#include <string.h>
#include <stdio.h>
#include "subject.h"

#define ARRAY_SIZE 25

using namespace std;

class Subject {
public:
	virtual ~Subject() { };

	virtual void Attach(Observer*);
	virtual void Detach(Observer*);
	virtual void Notify();
protected:
	Subject() { };

private:
	list<Observer*> _observers;
};

void Subject::Attach(Observer* o) {
	_observers.insert (_observers.end(), o);
	//Alternative syntax is _observers.push_back(o);
}

void Subject::Detach(Observer* o) {
	_observers.remove(o);
}

void Subject::Notify() {
	list<Observer*>::iterator i;

	for (i = _observers.begin(); i != _observers.end(); ++i)
	{
		Observer* myObserver = *i;
		myObserver->Update(this);
	}
}

/* OperationList is a concrete subject for maintaining the state
of a number of long-running operations. Examples of such operations are:
network device code/data backup/restore. Others are node/interface
state monitoring. The latter are often used to trigger failure
recovery mechanisms. */

class OperationList : public Subject {
public:
	OperationList ();
	void UpdateOperations(const char*);
	char* getOperationList();
private:
	char operationList[ARRAY_SIZE];
};

OperationList::OperationList() {
	memset(operationList, '\0', sizeof(operationList));
}

void OperationList::UpdateOperations(const char* updatedOp) {
	memcpy(operationList, updatedOp, strlen(updatedOp));
	Notify();
}

char* OperationList::getOperationList() {
	return operationList;
}

class NetOperationViewer : public Observer {
public:
	NetOperationViewer(OperationList*);
	virtual ~NetOperationViewer();

	virtual void Update(Subject*);
	virtual void Draw();

private:
	OperationList* _subject;
};

NetOperationViewer::NetOperationViewer(OperationList* s) {
	_subject = s;
	_subject->Attach(this);
	printf("Hello from observer\n");
}

NetOperationViewer::~NetOperationViewer() {
	_subject->Detach(this);
}

void NetOperationViewer::Update(Subject* changedSubject) {
	if (changedSubject == _subject)
		Draw();
}

void NetOperationViewer::Draw() {
	printf("Updating observer operation data %s\n", 
		_subject->getOperationList());
}

class NetOperationPercentViewer : public Observer {
public:
	NetOperationPercentViewer(OperationList*);
	virtual ~NetOperationPercentViewer();

	virtual void Update(Subject*);
	virtual void Draw();

private:
	OperationList* _subject;
};

NetOperationPercentViewer::NetOperationPercentViewer(OperationList* s) {
	_subject = s;
	_subject->Attach(this);
	printf("Hello from observer\n");
}

NetOperationPercentViewer::~NetOperationPercentViewer() {
	_subject->Detach(this);
}

void NetOperationPercentViewer::Update(Subject* changedSubject) {
	if (changedSubject == _subject)
		Draw();
}

void NetOperationPercentViewer::Draw() {
	char percentStr[4];
	char *buffer = _subject->getOperationList();

	memset(percentStr, '\0', sizeof(percentStr));
	strncpy(percentStr, buffer + 10, 3);
	
	printf("Updating observer operation data %s\n", percentStr);
}

INT _CRTAPI1 main(
    IN int  argumentCount,
    IN char *argumentVector[])
    {
		OperationList* myOperationList = new OperationList();
		NetOperationViewer* myNetOperationViewer = 
			new NetOperationViewer(myOperationList);
		
		NetOperationPercentViewer* myNetOperationViewer2 = 
			new NetOperationPercentViewer(myOperationList);

		//Now the asynchronous process updates the operation status
		myOperationList->UpdateOperations("DB Backup 10% Complete\0");

		//Now the asynchronous process updates the other operation status
		myOperationList->UpdateOperations("DB Backup 50% Complete\0");

		delete(myNetOperationViewer);
		delete(myOperationList);

		return 0;
}
